""" User State Get Controller Definition """

from starlette import status
from starlette.exceptions import HTTPException

from rowantree.contracts import (
    FeatureType,
    IncomeSourceType,
    StoreType,
    UserFeatureState,
    UserIncome,
    UserNotification,
    UserState,
    UserStore,
)

from ..services.db.incorrect_row_count_error import IncorrectRowCountError
from .abstract_controller import AbstractController


class UserStateGetController(AbstractController):
    """
    User State Get Controller
    Gets the user game state.

    Methods
    -------
    execute(self, user_guid: str) -> UserState
        Executes the command.
    """

    def execute(self, user_guid: str) -> UserState:
        """
        Gets the user game state.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_state: UserState
            The user state object.
        """

        try:
            # User Game State
            active: int = self.dao.user_active_state_get(user_guid=user_guid)

            # User Stores (Inventory)
            stores: dict[StoreType, UserStore] = self.dao.user_stores_get(user_guid=user_guid)

            # User Income
            incomes: dict[IncomeSourceType, UserIncome] = self.dao.user_income_get(user_guid=user_guid)

            # Features
            features: set[FeatureType] = self.dao.user_features_get(user_guid=user_guid)

            # Population
            population: int = self.dao.user_population_by_guid_get(user_guid=user_guid)

            # Active Feature Details and name
            active_feature_state: UserFeatureState = self.dao.user_active_feature_state_details_get(user_guid=user_guid)
            active_feature_state.name = self.dao.user_active_feature_get(user_guid=user_guid)

            # Merchants
            merchants: set[StoreType] = self.dao.user_merchant_transforms_get(user_guid=user_guid)

            # Notifications
            notifications: list[UserNotification] = self.dao.user_notifications_get(user_guid=user_guid)

        except IncorrectRowCountError as error:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="User not found") from error

        user_state: UserState = UserState(
            active=active,
            stores=stores,
            incomes=incomes,
            features=features,
            active_feature_state=active_feature_state,
            population=population,
            merchants=merchants,
            notifications=notifications,
        )
        return user_state



""" User Merchant Transforms Get Controller Definition """
from rowantree.contracts import StoreType
from rowantree.service.sdk import MerchantTransformsGetResponse

from .abstract_controller import AbstractController


class UserMerchantTransformsGetController(AbstractController):
    """
    User Merchant Transforms Get Controller
    Gets a (unique) list of user merchant transforms.

    Methods
    -------
    execute(self, user_guid: str) -> UserMerchants
        Executes the command.
    """

    def execute(self, user_guid: str) -> MerchantTransformsGetResponse:
        """
        Gets a (unique) list of user merchant transforms.

        Parameters
        ----------
        user_guid: str
            Target user guid.

        Returns
        -------
        user_merchants: MerchantTransformsGetResponse
            A (unique) list of user merchant transforms.
        """

        merchants: set[StoreType] = self.dao.user_merchant_transforms_get(user_guid=user_guid)
        return MerchantTransformsGetResponse(merchants=merchants)



""" User Delete Controller Definition """

from .abstract_controller import AbstractController


class UserDeleteController(AbstractController):
    """
    User Delete Controller
    Deletes a user.

    Methods
    -------
    execute(self, user_guid: str) -> None
        Executes the command.
    """

    def execute(self, user_guid: str) -> None:
        """
        Deletes a user.
        TODO: the underlying calls need to provide more context on status of this call.

        Parameters
        ----------
        user_guid: str
            The target user guid.
        """

        self.dao.user_delete(user_guid=user_guid)



""" DB DAO Utilities """

import logging
import socket
import time
from typing import Optional

from mysql.connector import Error, errorcode
from mysql.connector.pooling import MySQLConnectionPool

from rowantree.common.sdk import demand_env_var


# pylint: disable=duplicate-code
def get_connect_pool() -> MySQLConnectionPool:
    """
    Gets a connection pool to the database.

    Returns
    -------
    cnxpool: MySQLConnectionPool
        A connection pool to the database server.
    """

    try:
        logging.debug("Connecting to database")
        cnxpool: MySQLConnectionPool = MySQLConnectionPool(
            pool_name="servercnxpool",
            pool_size=3,
            user=demand_env_var(name="DATABASE_USERNAME"),
            password=demand_env_var(name="DATABASE_PASSWORD"),
            host=demand_env_var(name="DATABASE_SERVER"),
            database=demand_env_var(name="DATABASE_NAME"),
        )
        return cnxpool
    except socket.error as error:
        logging.debug(str(error))
        raise error from error
    except Error as error:
        logging.debug(str(error))
        if error.errno == errorcode.ER_ACCESS_DENIED_ERROR:
            logging.debug("Something is wrong with your user name or password")
        elif error.errno == errorcode.ER_BAD_DB_ERROR:
            logging.debug("Database does not exist")
        raise error from error
    except Exception as error:
        logging.debug(str(error))
        raise error from error


class WrappedConnectionPool:
    """
    Manages the creation of a MySQL connection pool along with reties.

    Attributes
    ----------
    retry_wait: int = 1
        The number of seconds to wait between attempts.
    healthy: bool = False
        Whether a pool was created.
    cnxpool: Optional[MySQLConnectionPool] = None
        The created connection pool.
    """

    retry_wait: int = 1
    healthy: bool = False
    cnxpool: Optional[MySQLConnectionPool] = None

    def __init__(self):
        # Attempts forever.  In the expected consumption scenario we would want this to occur indefinitely.
        while not self.healthy:
            self.healthy = self.connect()

    def connect(self) -> bool:
        """
        Attempts to create a database connection pool.

        Returns
        -------
        result: bool
            True or False depending on whether a pool was successfully created.
        """

        # pylint: disable=broad-except
        try:
            self.cnxpool = get_connect_pool()
            return True
        except Exception as error:
            logging.error("Unable to create connection pool, retrying ..")
            logging.error(str(error))
            time.sleep(self.retry_wait)
            return False



""" World Status Get Controller Definition """
from rowantree.service.sdk import WorldStatus

from ..controllers.abstract_controller import AbstractController


class WorldStatusGetController(AbstractController):
    """
    World Status Get Controller
    Gets the world status.

    Methods
    -------
    execute(self) -> WorldStatus
        Executes the command.
    """

    def execute(self) -> WorldStatus:
        """
        Gets the world status.

        Returns
        -------
        world_status: WorldStatus
            The world status.
        """

        return WorldStatus(active_users=self.dao.users_active_get())



""" User Income Set Controller Definition """

from rowantree.service.sdk import UserIncomeSetRequest

from ..controllers.abstract_controller import AbstractController


class UserIncomeSetController(AbstractController):
    """
    User Income Set Controller
    Sets a user income. (Creates or dismisses a number of workers of the type).

    Methods
    -------
    execute(self, user_guid: str, request: UserIncomeSetRequest) -> None
        Executes the command.
    """

    def execute(self, user_guid: str, request: UserIncomeSetRequest) -> None:
        """
        Sets a user income. (Creates or dismisses a number of workers of the type).

        Parameters
        ----------
        user_guid: str
            The target user guid.
        request: UserIncomeSetRequest
            The UserIncomeSetRequest object for the update.
        """

        self.dao.user_income_set(user_guid=user_guid, transaction=request)



""" Merchant Transform Perform Controller Definition """

from rowantree.service.sdk import MerchantTransformRequest

from .abstract_controller import AbstractController


class MerchantTransformPerformController(AbstractController):
    """
    Merchant Transform Perform Controller

    Methods
    -------
    execute(self, user_guid: str, request: MerchantTransformRequest) -> None
        Executes the command.
    """

    def execute(self, user_guid: str, request: MerchantTransformRequest) -> None:
        """
        Performs a merchant transform.

        Parameters
        ----------
        user_guid: str
            The target user guid.
        request: MerchantTransformRequest
            The merchant transform request.
        """

        self.dao.merchant_transform_perform(user_guid=user_guid, store_name=request.store_name)



""" User Features Get Controller Definition """
import logging

from starlette import status
from starlette.exceptions import HTTPException

from rowantree.contracts import FeatureType
from rowantree.service.sdk import FeaturesGetResponse

from ..services.db.incorrect_row_count_error import IncorrectRowCountError
from .abstract_controller import AbstractController


class UserFeaturesGetController(AbstractController):
    """
    User Features Get Controller
    Gets the unique list of user features.

    Methods
    -------
    execute(self, user_guid: str) -> UserFeatures
        Executes the command.
    """

    def execute(self, user_guid: str) -> FeaturesGetResponse:
        """
        Gets the unique list of user features.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_features: UserFeatures
            A unique list of user features.
        """

        try:
            features: set[FeatureType] = self.dao.user_features_get(user_guid=user_guid)
            return FeaturesGetResponse(features=features)
        except IncorrectRowCountError as error:
            # User did not exist (received an empty tuple)
            logging.debug("caught: {%s}", str(error))
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Unable to find user") from error



""" User Income Get Controller Definition """
from rowantree.contracts import IncomeSourceType, UserIncome
from rowantree.service.sdk import UserIncomeGetResponse

from .abstract_controller import AbstractController


class UserIncomeGetController(AbstractController):
    """
    User Income Get Controller
    Gets (unique) list of user incomes.

    Methods
    -------
    execute(self, user_guid: str) -> UserIncomeGetResponse
        Executes the command.
    """

    def execute(self, user_guid: str) -> UserIncomeGetResponse:
        """
        Gets (unique) list of user incomes.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_incomes: UserIncomeGetResponse
            A (unique) list of user incomes.
        """

        income_sources: dict[IncomeSourceType, UserIncome] = self.dao.user_income_get(user_guid=user_guid)
        return UserIncomeGetResponse(incomes=income_sources)



""" The Rowan Tree Service Layer Handler(s) """

import logging
import os
from pathlib import Path

from fastapi import Depends, FastAPI, status
from starlette.exceptions import HTTPException
from starlette.middleware.cors import CORSMiddleware

from rowantree.auth.sdk.common.depends import is_admin, is_enabled
from rowantree.auth.sdk.contracts.dto.token_claims import TokenClaims
from rowantree.common.sdk import demand_env_var
from rowantree.contracts import ActionQueue, User, UserFeatureState, UserState
from rowantree.service.sdk import (
    FeaturesGetResponse,
    MerchantTransformRequest,
    MerchantTransformsGetResponse,
    PopulationGetResponse,
    StoresGetResponse,
    UserActiveGetStatus,
    UserIncomeGetResponse,
    UserIncomeSetRequest,
    UserTransportRequest,
    WorldStatus,
)

from ..controllers.action_queue_process import ActionQueueProcessController
from ..controllers.merchant_transforms_perform import MerchantTransformPerformController
from ..controllers.user_active_get import UserActiveGetController
from ..controllers.user_active_set import UserActiveSetController
from ..controllers.user_create import UserCreateController
from ..controllers.user_delete import UserDeleteController
from ..controllers.user_feature_active_get import UserFeatureActiveGetController
from ..controllers.user_features_get import UserFeaturesGetController
from ..controllers.user_income_get import UserIncomeGetController
from ..controllers.user_income_set import UserIncomeSetController
from ..controllers.user_merchant_transforms_get import UserMerchantTransformsGetController
from ..controllers.user_population_get import UserPopulationGetController
from ..controllers.user_state_get import UserStateGetController
from ..controllers.user_stores_get import UserStoresGetController
from ..controllers.user_transport import UserTransportController
from ..controllers.world_get import WorldStatusGetController
from ..services.db.dao import DBDAO
from ..services.db.utils import WrappedConnectionPool

# Setup logging
Path(demand_env_var(name="LOGS_DIR")).mkdir(parents=True, exist_ok=True)
logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
    level=logging.DEBUG,
    filemode="w",
    filename=f"{demand_env_var(name='LOGS_DIR')}/{os.uname()[1]}.therowantree.service.log",
)

logging.debug("Starting server")

# Creating database connection pool, and DAO
wrapped_cnxpool: WrappedConnectionPool = WrappedConnectionPool()
dao: DBDAO = DBDAO(cnxpool=wrapped_cnxpool.cnxpool)

# Create controllers
merchant_transforms_perform_controller = MerchantTransformPerformController(dao=dao)
user_active_get_controller = UserActiveGetController(dao=dao)
user_active_set_controller = UserActiveSetController(dao=dao)
user_create_controller = UserCreateController(dao=dao)
user_delete_controller = UserDeleteController(dao=dao)
user_features_get_controller = UserFeaturesGetController(dao=dao)
user_features_active_get_controller = UserFeatureActiveGetController(dao=dao)
user_income_get_controller = UserIncomeGetController(dao=dao)
user_income_set_controller = UserIncomeSetController(dao=dao)
user_merchant_transforms_get_controller = UserMerchantTransformsGetController(dao=dao)
user_population_get_controller = UserPopulationGetController(dao=dao)
user_transport_controller = UserTransportController(dao=dao)
user_state_get_controller = UserStateGetController(dao=dao)
user_stores_get_controller = UserStoresGetController(dao=dao)
world_status_get_controller = WorldStatusGetController(dao=dao)
action_queue_process_controller = ActionQueueProcessController(dao=dao)

# Create the FastAPI application
app = FastAPI()

#  Apply COR Configuration | https://fastapi.tiangolo.com/tutorial/cors/
origins = ["*"]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Define our handlers
# pylint: disable=unused-argument


@app.get("/health/plain", status_code=status.HTTP_200_OK)
async def health_plain() -> bool:
    """
    Get Application Health
    [GET] /health/plain

    Returns
    -------
    [STATUS CODE] 200: OK
        health: bool
            A true/false response of server health.
    """

    return True


@app.post("/v1/user/{user_guid}/merchant", status_code=status.HTTP_201_CREATED)
def merchant_transforms_perform_handler(
    user_guid: str, request: MerchantTransformRequest, token_claims: TokenClaims = Depends(is_enabled)
) -> None:
    """
    Perform Merchant Exchange (Transform) For User
    [POST] /v1/user/{user_guid}/merchant

    Path
    ----
    user_guid: str
        The target user guid.

    Body
    ----
    request: MerchantTransformRequest
        The requested merchant transform.

    Returns
    -------
    [STATUS CODE]
        201 - Accepted
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        merchant_transforms_perform_controller.execute(user_guid=user_guid, request=request)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.get("/v1/user/{user_guid}/merchant", status_code=status.HTTP_200_OK)
def user_merchant_transforms_get_handler(
    user_guid: str, token_claims: TokenClaims = Depends(is_enabled)
) -> MerchantTransformsGetResponse:
    """
    Gets user merchants.
    [GET] /v1/user/{user_guid}/merchant

    Path
    ----------
    user_guid: str
        The target user guid.

    Returns
    -------
    user_merchants: MerchantTransformsGetResponse
        The user merchants.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_merchant_transforms_get_controller.execute(user_guid=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


# Get User's Active State
@app.get("/v1/user/{user_guid}/active", status_code=status.HTTP_200_OK)
def user_active_get_handler(user_guid: str, token_claims: TokenClaims = Depends(is_enabled)) -> UserActiveGetStatus:
    """
    Gets user's active state.
    [GET] /v1/user/{user_guid}/active

    Path
    ----------
    user_guid: str
        The target user guid.

    Returns
    -------
    user_active_state: UserActiveGetStatus
        The user active state

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_active_get_controller.execute(user_guid=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


# Set User's Active State
@app.post("/v1/user/{user_guid}/active", status_code=status.HTTP_200_OK)
def user_active_set_handler(
    user_guid: str, request: UserActiveGetStatus, token_claims: TokenClaims = Depends(is_enabled)
) -> None:
    """
    Sets user's active state.
    [POST] /v1/user/{user_guid}/active

    Path
    ----------
    user_guid: str
        The target user guid.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        user_active_set_controller.execute(user_guid=user_guid, request=request)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


# Create User
@app.post("/v1/user/{user_guid}", status_code=status.HTTP_201_CREATED)
def user_create_handler(user_guid: str, token_claims: TokenClaims = Depends(is_enabled)) -> User:
    """
    Creates a user.
    [POST] /v1/user

    Returns
    -------
    user_active_state: User
        The user.

    [STATUS CODE]
        201 - Accepted
        401 - Not authorized
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_create_controller.execute(request=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


# Delete User
@app.delete("/v1/user/{user_guid}", status_code=status.HTTP_200_OK)
def user_delete_handler(user_guid: str, token_claims: TokenClaims = Depends(is_enabled)) -> None:
    """
    Deletes a user.
    [DELETE] /v1/user/{user_guid}

    Path
    ----------
    user_guid: str
        The target user guid.

    Returns
    -------
    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        user_delete_controller.execute(user_guid=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.get("/v1/user/{user_guid}/features", status_code=status.HTTP_200_OK)
def user_features_get_handler(user_guid: str, token_claims: TokenClaims = Depends(is_enabled)) -> FeaturesGetResponse:
    """
    Get User Features.
    [GET] /v1/user/{user_guid}/features

    Path
    ----------
    user_guid: str
        The target user guid.

    Returns
    -------
    user_features: FeaturesGetResponse
    The user features.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_features_get_controller.execute(user_guid=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.get("/v1/user/{user_guid}/features/active", status_code=status.HTTP_200_OK)
def user_feature_active_get_handler(
    user_guid: str, token_claims: TokenClaims = Depends(is_enabled)
) -> UserFeatureState:
    """
    Get Active User Features.
    [GET] /v1/user/{user_guid}/features/active

    Path
    ----------
    user_guid: str
        The target user guid.

    Returns
    -------
    user_feature: UserFeatureState
    The active user feature state.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_features_active_get_controller.execute(user_guid=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.get("/v1/user/{user_guid}/income", status_code=status.HTTP_200_OK)
def user_income_get_handler(user_guid: str, token_claims: TokenClaims = Depends(is_enabled)) -> UserIncomeGetResponse:
    """
    Get User Income Sources.
    [GET] /v1/user/{user_guid}/income

    Path
    ----------
    user_guid: str
        The target user guid.

    Returns
    -------
    user_incomes: UserIncomeGetResponse
    The user income sources.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_income_get_controller.execute(user_guid=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.post("/v1/user/{user_guid}/income", status_code=status.HTTP_200_OK)
def user_income_set_handler(
    user_guid: str, request: UserIncomeSetRequest, token_claims: TokenClaims = Depends(is_enabled)
) -> None:
    """
    Set User Income Source.
    [POST] /v1/user/{user_guid}/income

    Path
    ----------
    user_guid: str
        The target user guid.

    Body
    ----
    request: UserIncomeSetRequest
        The user income set request.

    Returns
    -------
    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        user_income_set_controller.execute(user_guid=user_guid, request=request)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.get("/v1/user/{user_guid}/population", status_code=status.HTTP_200_OK)
def user_population_get_handler(
    user_guid: str, token_claims: TokenClaims = Depends(is_enabled)
) -> PopulationGetResponse:
    """
    Set User Population.
    [GET] /v1/user/{user_guid}/population

    Path
    ----------
    user_guid: str
        The target user guid.

    Returns
    -------
    user_population: PopulationGetResponse
        User population.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_population_get_controller.execute(user_guid=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.post("/v1/user/{user_guid}/transport", status_code=status.HTTP_200_OK)
def user_transport_handler(
    user_guid: str, request: UserTransportRequest, token_claims: TokenClaims = Depends(is_enabled)
) -> UserFeatureState:
    """
    Transport User
    [POST] /v1/user/{user_guid}/transport

    Path
    ----------
    user_guid: str
        The target user guid.

    Body
    ----
    request: UserTransportRequest
        The user transport request.

    Returns
    -------
    active_user_feature_state: UserFeatureState
        Active User Feature State.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_transport_controller.execute(user_guid=user_guid, request=request)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.get("/v1/user/{user_guid}/state", status_code=status.HTTP_200_OK)
def user_state_get_handler(user_guid: str, token_claims: TokenClaims = Depends(is_enabled)) -> UserState:
    """
    Get User State
    [GET] /v1/user/{user_guid}/state

    Path
    ----------
    user_guid: str
        The target user guid.

    Returns
    -------
    user_state: UserState
        User state.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_state_get_controller.execute(user_guid=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.get("/v1/user/{user_guid}/stores", status_code=status.HTTP_200_OK)
def user_stores_get_handler(user_guid: str, token_claims: TokenClaims = Depends(is_enabled)) -> StoresGetResponse:
    """
    Get User Stores
    [GET] /v1/user/{user_guid}/stores

    Path
    ----------
    user_guid: str
        The target user guid.

    Returns
    -------
    user_stores: StoresGetResponse
        User stores.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        404 - User not found
        500 - Server failure
    """

    try:
        if user_guid != token_claims.sub and not token_claims.admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
            )

        return user_stores_get_controller.execute(user_guid=user_guid)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.get("/v1/world", status_code=status.HTTP_200_OK)
def world_get_handler(token_claims: TokenClaims = Depends(is_admin)) -> WorldStatus:
    """
    Get World Status
    [GET] /v1/world

    Returns
    -------
    world_status: WorldStatus
        World status.

    [STATUS CODE]
        200 - OK
        401 - Not authorized
        500 - Server failure
    """

    try:
        return world_status_get_controller.execute()
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error


@app.post("/v1/world/queue", status_code=status.HTTP_200_OK)
def action_queue_process_handler(request: ActionQueue, token_claims: TokenClaims = Depends(is_admin)) -> None:
    """
    Progress Action Queue
    [POST] /v1/world/queue

    Body
    ----
    request: ActionQueue
        Action queue.

    Returns
    -------
    [STATUS CODE]
        200 - OK
        401 - Not authorized
        500 - Server failure
    """

    try:
        action_queue_process_controller.execute(request=request)
    except HTTPException as error:
        logging.error(str(error))
        raise error from error
    except Exception as error:
        # Caught all other uncaught errors.
        logging.error(str(error))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
        ) from error



""" Action Queue Process Controller Definition """

from rowantree.contracts import ActionQueue

from .abstract_controller import AbstractController


class ActionQueueProcessController(AbstractController):
    """
    Action Queue Process Controller
    Processes the requested action queue.

    Methods
    -------
    execute(self, action_queue: ActionQueue) -> None
        Execute the controller.
    """

    def execute(self, request: ActionQueue) -> None:
        """
        Processes the requested action queue.

        Parameters
        ----------
        request: ActionQueue
            The process action queue.
        """

        self.dao.process_action_queue(action_queue=request)



""" User Feature Active Get Controller Definition """
import logging

from starlette import status
from starlette.exceptions import HTTPException

from rowantree.contracts import UserFeatureState

from ..services.db.incorrect_row_count_error import IncorrectRowCountError
from .abstract_controller import AbstractController


class UserFeatureActiveGetController(AbstractController):
    """
    User Feature Active Get Controller
    Gets the active user feature.

    Methods
    -------
    execute(self, user_guid: str, details: bool) -> UserFeature
        Executes the command.
    """

    def execute(self, user_guid: str) -> UserFeatureState:
        """
        Gets the active user feature.

        Parameters
        ----------
        user_guid: str
            The target user guid.
        details: bool
            Whether to include details of the feature.

        Returns
        -------
        user_feature: UserFeature
            The active UserFeature.
        """

        try:
            return self.dao.user_active_feature_get(user_guid=user_guid)
        except IncorrectRowCountError as error:
            # User did not exist (received an empty tuple)
            logging.debug("caught: {%s}", str(error))
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Unable to find user") from error



""" User Transport Controller Definition """
import logging

from starlette import status
from starlette.exceptions import HTTPException

from rowantree.contracts import UserFeatureState
from rowantree.service.sdk import UserTransportRequest

from ..contracts.sql_exception_error import SqlExceptionError
from ..services.db.incorrect_row_count_error import IncorrectRowCountError
from .abstract_controller import AbstractController


class UserTransportController(AbstractController):
    """
    User Transport Controller
    Performs a user transport. (feature to feature change)

    Methods
    -------
    execute(self, user_guid: str, request: UserTransportRequest) -> UserFeatureState
        Executes the command.
    """

    def execute(self, user_guid: str, request: UserTransportRequest) -> UserFeatureState:
        """
        Performs a user transport. (feature to feature change)

        Parameters
        ----------
        user_guid: str
            The target user guid.
        request: UserTransportRequest
            The UserTransportRequest to perform.

        Returns
        -------
        active_feature_state: UserFeatureState
            The user's new active feature state.
        """

        try:
            return self.dao.user_transport(user_guid=user_guid, location=request.location)
        except IncorrectRowCountError as error:
            logging.debug("caught: {%s}", str(error))
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Unable to find user") from error
        except SqlExceptionError as error:
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Unable move to location") from error



""" User Create Controller Definition """


from starlette import status
from starlette.exceptions import HTTPException

from rowantree.contracts import User

from ..services.db.incorrect_row_count_error import IncorrectRowCountError
from .abstract_controller import AbstractController


class UserCreateController(AbstractController):
    """
    User Create Controller
    Creates a user.

    Methods
    -------
    execute(self) -> User
        Executes the command.
    """

    def execute(self, request: str) -> User:
        """
        Creates a user.

        Returns
        -------
        user: User
            The newly created user.
        """

        try:
            return self.dao.user_create_by_guid(user_guid=request)
        except IncorrectRowCountError as error:
            raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail="Unable to create new user") from error



""" User Active Get Controller Definition """

import logging

from starlette import status
from starlette.exceptions import HTTPException

from rowantree.service.sdk import UserActiveGetStatus

from ..services.db.incorrect_row_count_error import IncorrectRowCountError
from .abstract_controller import AbstractController


class UserActiveGetController(AbstractController):
    """
    User Active Get Controller
    Gets the user active state.
    """

    def execute(self, user_guid: str) -> UserActiveGetStatus:
        """
        Gets the user active state.
        If the requested user does not exist we do not expose this in the response. (information leakage).
        If the user is not found or is inactive we return an inactive response.

        Parameters
        ----------
        user_guid: str
            The user guid to look up.

        Returns
        -------
        user_active: UserActiveGetStatus
            The user active state object.
        """

        try:
            active: bool = self.dao.user_active_state_get(user_guid=user_guid)
            return UserActiveGetStatus(active=active)
        except IncorrectRowCountError as error:
            logging.debug("caught: {%s}", str(error))
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Unable to find user") from error



""" Abstract Controller Definition """

from abc import abstractmethod
from typing import Any, Optional

from rowantree.contracts import BaseModel

from ..services.db.dao import DBDAO


class AbstractController(BaseModel):
    """
    Abstract Controller

    Attributes
    ----------
    dao: DBDAO
        The database DAO.
    """

    dao: DBDAO

    @abstractmethod
    def execute(self, *args, **kwargs) -> Optional[Any]:
        """Should be implemented in the subclass"""



""" User Active Set Controller Definition """

from rowantree.service.sdk import UserActiveGetStatus

from .abstract_controller import AbstractController


class UserActiveSetController(AbstractController):
    """
    User Active Set Controller
    Sets the user active state.

    Methods
    -------
    execute(self, user_guid: str, request: UserActive) -> UserActive
        Executes the command.
    """

    def execute(self, user_guid: str, request: UserActiveGetStatus) -> None:
        """
        Sets the user active state.

        Parameters
        ----------
        user_guid: str
            The user guid to target.
        request: UserActiveGetStatus
            The active state to set the user to.
        """

        # This is atomic and idempotent.  If it failed then we most likely are running into
        # either database level connectivity, config, or table leve locks...
        # TODO: the underlying calls need to provide more context on status of this call.

        self.dao.user_active_state_set(user_guid=user_guid, active=request.active)



""" User Stores Get Controller Definition """
from rowantree.contracts import StoreType, UserStore
from rowantree.service.sdk import StoresGetResponse

from .abstract_controller import AbstractController


class UserStoresGetController(AbstractController):
    """
    User Stores Get Controller
    Gets the (unique) list of user stores.

    Methods
    -------
    execute(self, user_guid: str) -> UserStores
        Executes the command.
    """

    def execute(self, user_guid: str) -> StoresGetResponse:
        """
        Gets the (unique) list of user stores.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_stores: UserStores
            A (unique) list of user stores.
        """

        stores: dict[StoreType, UserStore] = self.dao.user_stores_get(user_guid=user_guid)
        return StoresGetResponse(stores=stores)



""" Database DAO Definition """

import logging
import socket
from datetime import datetime
from typing import Optional, Tuple

import mysql.connector
from mysql.connector import IntegrityError, errorcode
from mysql.connector.pooling import MySQLConnectionPool, PooledMySQLConnection
from starlette import status
from starlette.exceptions import HTTPException

from rowantree.contracts import (
    ActionQueue,
    BaseModel,
    FeatureDetailType,
    FeatureType,
    IncomeSourceType,
    StoreType,
    User,
    UserEvent,
    UserFeatureState,
    UserIncome,
    UserNotification,
    UserStore,
)
from rowantree.service.sdk import UserIncomeSetRequest

from ...contracts.duplicate_key_error import DuplicateKeyError
from ...contracts.sql_exception_error import SqlExceptionError
from .incorrect_row_count_error import IncorrectRowCountError


class DBDAO(BaseModel):
    """
    Database DAO

    Attributes
    ----------
    cnxpool: MySQLConnectionPool
        MySQL Connection Pool
    """

    cnxpool: MySQLConnectionPool

    def merchant_transform_perform(self, user_guid: str, store_name: str) -> None:
        """
        Perform a merchant transform.

        Parameters
        ----------
        user_guid: str
            The target user guid.
        store_name: str
            The name of the store to perform the transform on.
        """

        args: list = [user_guid, store_name]
        self._call_proc("peformMerchantTransformByGUID", args)

    def user_active_feature_get(self, user_guid: str) -> FeatureType:
        """
        Gets user's active feature/location.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_feature: FeatureType
            The active user feature.
        """

        args: list = [
            user_guid,
        ]
        rows: list[Tuple[str]] = self._call_proc("getUserActiveFeatureByGUID", args)
        if len(rows) != 1:
            # User did not exist (received an empty tuple)
            message: str = f"Result count was not exactly one. Received: {rows}"
            logging.debug(message)
            raise IncorrectRowCountError(message)
        return FeatureType(rows[0][0])

    def user_active_feature_state_details_get(self, user_guid: str) -> UserFeatureState:
        """
        Get User Active Feature/Location Including Details.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_feature: UserFeatureState
            The active user feature/location with details.
        """

        args: list = [
            user_guid,
        ]
        rows: list[Tuple[str, Optional[str]]] = self._call_proc("getUserActiveFeatureStateDetailsByGUID", args)
        if len(rows) != 1:
            # User did not exist (received an empty tuple)
            message: str = f"Result count was not exactly one. Received: {rows}"
            logging.debug(message)
            raise IncorrectRowCountError(message)

        return UserFeatureState(details=FeatureDetailType(rows[0][0]), description=rows[0][1])

    def users_active_get(self) -> set[str]:
        """
        Get Active Users.

        Returns
        -------
        active_user_guids: set[str]
            A (unique) set of user guids which are active.
        """

        active_user_guids: set[str] = set()
        rows: list[Tuple] = self._call_proc("getActiveUsers", [])
        for response_tuple in rows:
            active_user_guids.add(response_tuple[0])
        return active_user_guids

    def user_active_state_get(self, user_guid: str) -> bool:
        """
        Get user active state.

        Parameters
        ----------
        user_guid: str
        The target user guid.

        Returns
        -------
        user_active: UserActive
            The user active state.
        """

        args: list[str, int] = [
            user_guid,
        ]
        rows: list[Tuple[int]] = self._call_proc("getUserActivityStateByGUID", args)
        if len(rows) != 1:
            raise IncorrectRowCountError(f"Result count was not exactly one. Received: {rows}")
        if rows[0][0] == 1:
            return True
        return False

    def user_active_state_set(self, user_guid: str, active: bool) -> None:
        """
        Set user's active state.
        # TODO: the underlying calls need to provide more context on status of this call.

        Parameters
        ----------
        user_guid: str
            The target user guid.
        active: bool
            The active state to set.
        """

        args: list = [
            user_guid,
        ]
        if active:
            proc = "setUserActiveByGUID"
        else:
            proc = "setUserInactiveByGUID"
        self._call_proc(name=proc, args=args)

    def user_create_by_guid(self, user_guid: str) -> User:
        """
        Create a user.
        TODO this returns nothing, needs more detail from the db.

        Returns
        -------
        user: User
            The created user.
        """

        args = [user_guid]
        try:
            self._call_proc("createUserByGUID", args)
        except (IntegrityError, DuplicateKeyError) as error:
            message: str = f"User already exists: {user_guid}, {str(error)}"
            logging.debug(message)
            raise IncorrectRowCountError(message) from error
        return User(guid=user_guid)

    def user_delete(self, user_guid: str) -> None:
        """
        Delete user.
        TODO: the underlying calls need to provide more context on status of this call.

        Parameters
        ----------
        user_guid: str
            The target user guid.
        """

        args: list = [
            user_guid,
        ]
        self._call_proc("deleteUserByGUID", args)

    def user_features_get(self, user_guid: str) -> set[FeatureType]:
        """
        Get user features.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_features: UserFeatures
            User features object.
        """

        features: set[FeatureType] = set()

        args: list = [
            user_guid,
        ]
        rows: list[Tuple[str]] = self._call_proc("getUserFeaturesByGUID", args)

        if rows is None or len(rows) == 0:
            # User does not exist ot had no features.  However, it should not be
            # possible to exist in the game world with without a feature state.
            # Users are given a feature state on creation (the process is atomic).
            # This should always be do to a non-existent user.
            message: str = f"Result count was not exactly one. Received: {rows}"
            logging.debug(message)
            raise IncorrectRowCountError(message)

        for row in rows:
            name: str = row[0]
            features.add(FeatureType(name))

        return features

    def user_income_get(self, user_guid: str) -> dict[IncomeSourceType, UserIncome]:
        """
        Get user income sources
        # TODO: This can not tell if there are no stores, or if the user just doesn't exist.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_incomes: dict[IncomeSourceType, UserIncome]
        """

        income_sources: dict[IncomeSourceType, UserIncome] = {}

        args: list[str] = [
            user_guid,
        ]
        rows: list[Tuple[int, str, Optional[str]]] = self._call_proc("getUserIncomeByGUID", args)
        for row in rows:
            income_sources[IncomeSourceType(row[1])] = UserIncome(
                amount=row[0], name=IncomeSourceType(row[1]), description=row[2]
            )
        return income_sources

    def user_income_set(self, user_guid: str, transaction: UserIncomeSetRequest) -> None:
        """
        Set user income source.
        TODO: This needs to return more detail on success / failure.

        Parameters
        ----------
        user_guid: str
            The target user guid.
        transaction: UserIncomeSetRequest
            The user income set request.
        """

        args = [user_guid, transaction.income_source_name, transaction.amount]
        self._call_proc("deltaUserIncomeByNameAndGUID", args)

    def user_merchant_transforms_get(self, user_guid: str) -> set[StoreType]:
        """
        Get User merchant transforms [currently available].

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_merchants: set[StoreType]
            User merchants object.
        """

        merchants: set[StoreType] = set()

        args: list = [
            user_guid,
        ]
        rows: list[Tuple[str]] = self._call_proc("getUserMerchantTransformsByGUID", args)
        for row in rows:
            merchants.add(StoreType(row[0]))
        return merchants

    def user_notifications_get(self, user_guid: str) -> list[UserNotification]:
        """
        Get User notifications.
        Returns the most recent (new) user notifications.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_notifications: UserNotifications
            New user notifications.
        """

        notifications: list[UserNotification] = []

        args: list = [
            user_guid,
        ]
        rows: list[Tuple[int, datetime, str]] = self._call_proc("getUserNotificationByGUID", args)
        for row in rows:
            notification: UserNotification = UserNotification(
                index=row[0], timestamp=row[1], event=UserEvent.parse_raw(row[2])
            )
            notifications.append(notification)
        return notifications

    def user_population_by_guid_get(self, user_guid: str) -> int:
        """
        Get user population (by GUID)

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_population: int
            User population size.
        """

        rows: list[Tuple[int]] = self._call_proc(
            "getUserPopulationByGUID",
            [
                user_guid,
            ],
        )
        if rows is None or len(rows) != 1:
            message: str = f"Result count was not exactly one. Received: {rows}"
            logging.debug(message)
            raise IncorrectRowCountError(message)

        return rows[0][0]

    def user_stores_get(self, user_guid: str) -> dict[StoreType, UserStore]:
        """
        Get user stores.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_stores: dict[StoreType, UserStore]
            User Stores Object
        """

        stores: dict[StoreType, UserStore] = {}

        args: list[str, int] = [
            user_guid,
        ]
        rows: list[Tuple[str, Optional[str], int]] = self._call_proc("getUserStoresByGUID", args)
        for row in rows:
            stores[StoreType(row[0])] = UserStore(name=StoreType(row[0]), description=row[1], amount=row[2])
        return stores

    def user_transport(self, user_guid: str, location: str) -> UserFeatureState:
        """
        Perform User Transport.

        Parameters
        ----------
        user_guid: str
            The target user guid.
        location: str
            The name of the feature/location to transport the user to.

        Returns
        -------
        user_feature: UserFeatureState
            The active user feature state.
        """

        args: list = [user_guid, location]
        rows: list[Tuple[str, Optional[str]]] = self._call_proc("transportUserByGUID", args, True)
        if len(rows) != 1:
            # User did not exist (received an empty tuple)
            message: str = f"Result count was not exactly one. Received: {rows}"
            logging.debug(message)
            raise IncorrectRowCountError(message)
        return UserFeatureState(details=rows[0][0], description=rows[0][1])

    # Utility functions

    def process_action_queue(self, action_queue: ActionQueue) -> None:
        """
        Process the provided action queue.

        Parameters
        ----------
        action_queue: ActionQueue
            The action queue to process.
        """

        for action in action_queue.queue:
            self._call_proc(action.name, action.arguments)

    # pylint: disable=duplicate-code
    def _call_proc(self, name: str, args: list, debug: bool = False) -> Optional[list[Tuple]]:
        """
        Perform a stored procedure call.

        Parameters
        ----------
        name: str
            The name of the stored procedure to call.
        args: list
            The arguments to pass to the stored procedure.
        debug: bool
            Whether to log debug details about the call.

        Returns
        -------
        results: Optional[list[Tuple]]
            An optional list of tuples (rows) from the call.
        """

        if debug:
            logging.debug("[DAO] [Stored Proc Call Details] Name: {%s}, Arguments: {%s}", name, args)
        rows: Optional[list[Tuple]] = None
        cnx: Optional[PooledMySQLConnection] = None
        try:
            cnx = self.cnxpool.get_connection()
            cursor = cnx.cursor()
            cursor.callproc(name, args)
            for result in cursor.stored_results():
                rows = result.fetchall()
            cursor.close()
            cnx.close()
        except socket.error as error:
            logging.error("socket.error")
            logging.error(error)
            if cnx:
                cnx.close()
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
            ) from error
        except mysql.connector.Error as error:
            logging.error("mysql.connector.Error: {%i}", error.errno)
            logging.error(str(error))
            if cnx:
                cnx.close()
            # pylint: disable=no-else-raise
            if error.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                logging.error("Something is wrong with your user name or password")
                raise SqlExceptionError("Bad username or password") from error
            elif error.errno == errorcode.ER_BAD_DB_ERROR:
                logging.error("Database does not exist")
                raise SqlExceptionError("Database does not exist") from error
            elif error.errno == errorcode.ER_DUP_ENTRY:
                logging.error("Duplicate key")
                raise DuplicateKeyError(str(error)) from error
            raise SqlExceptionError(f"Unhandled SQL Exception: {str(error)}") from error
        except Exception as error:
            logging.error("error")
            # All other uncaught exception types
            logging.error(str(error))
            if cnx:
                cnx.close()
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Internal Server Error"
            ) from error

        if debug:
            logging.debug("[DAO] [Stored Proc Call Details] Returning:")
            logging.debug(rows)
        return rows



""" User Population Get Controller Definition """
import logging

from starlette import status
from starlette.exceptions import HTTPException

from rowantree.service.sdk import PopulationGetResponse

from ..services.db.incorrect_row_count_error import IncorrectRowCountError
from .abstract_controller import AbstractController


class UserPopulationGetController(AbstractController):
    """
    User Population Get Controller
    Gets the user population.

    Methods
    -------
    execute(self, user_guid: str) -> UserPopulation
        Executes the command.
    """

    def execute(self, user_guid: str) -> PopulationGetResponse:
        """
        Gets the user population.

        Parameters
        ----------
        user_guid: str
            The target user guid.

        Returns
        -------
        user_population: UserPopulation
            User population object.
        """

        try:
            population: int = self.dao.user_population_by_guid_get(user_guid=user_guid)
            return PopulationGetResponse(population=population)
        except IncorrectRowCountError as error:
            logging.debug("caught: {%s}", str(error))
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Unable to find user") from error
